// INT17XXX.H:  saving and restoring registers during interrupt
/*
 IMPORTANT: CC7A will AUTOMATICALLY check that vital registers are
 saved and restored during interrupt. This applies to:

   Group 1:  W/WREG, ALUSTA, PCLATH, BSR  : most frequent used
   Group 2:  FSR0, FSR1                   : indirect access
   Group 3:  TBLPTR                       : reading 'const' data
   Group 4:  PRODL, PRODH                 : multiplication instructions

 NOTE that it is not required to save registers before starting to
 service the interrupt. Section 'Custom interrupt save and restore',
 shows a list of instructions that that will not disturb the main
 registers.

 It is normally not required to save PCLATH on devices using one code
 page only (max 8192 words of code). However, PCLATH must be saved
 if computed goto is used in both the main() routine and interrupt.

 It is possible to limit the save and restore of a specific register
 to a small region inside the interrupt service routine, if this
 register is modified only inside this region.

 CC7A supports CUSTOM save and restore sequences. If you want to use
 your own register save and restore during interrupt, please read
 the following CUSTOM INTERRUPT SAVE AND RESTORE section.


// DEFAULT INTERRUPT STRUCTURE (RECOMMENDED)

   #include "int17XXX.h"

   #pragma origin = 0x8
   interrupt int_server_( void)
   {
       multi_interrupt_entry_and_save


    PERIPHERAL_service:
        // save on demand: PRODL,PRODH,TBLPTRH,TBLPTRL,FSR0,FSR1
        /* process periferal interrupt */
        // the peripheral interrupt flag must be cleared manually
        // restore on demand: PRODL,PRODH,TBLPTRH,TBLPTRL,FSR0,FSR1

        goto RESTORE_and_return;


    TMR0_service:
        // save on demand: PRODL,PRODH,TBLPTRH,TBLPTRL,FSR0,FSR1
        /* process Timer 0 interrupt */
        // T0IF is automatically cleared when the CPU vectors to 0x10
        // restore on demand: PRODL,PRODH,TBLPTRH,TBLPTRL,FSR0,FSR1

        goto RESTORE_and_return;


    T0CKI_service:
        // save on demand: PRODL,PRODH,TBLPTRH,TBLPTRL,FSR0,FSR1
        /* process T0CKI pin interrupt */
        // T0CKIF is automatically cleared when the CPU vectors to 0x18
        // restore on demand: PRODL,PRODH,TBLPTRH,TBLPTRL,FSR0,FSR1

        goto RESTORE_and_return;


    INT_service:
        // save on demand: PRODL,PRODH,TBLPTRH,TBLPTRL,FSR0,FSR1
        /* process INT pin interrupt */
        // INTF is automatically cleared when the CPU vectors to 0x8
        // restore on demand: PRODL,PRODH,TBLPTRH,TBLPTRL,FSR0,FSR1

      RESTORE_and_return:
       interrupt_exit_and_restore
   }

   /* IMPORTANT : GLINTD should normally NOT be cleared in the
      interrupt routine. GLINTD is AUTOMATICALLY set on interrupt
      entry by the CPU and cleared on exit (by RETFIE). Clearing
      GLINTD inside the interrupt service routine will cause nested
      interrupts if an interrupt is pending. This is incompatible
      with the RAM allocation for local variables, and care must be
      taken to ensure that the RAM locations (for context save and
      restore) are not overwritten by a nested interrupt. Also,
      remember that too deep nesting may crash the program! CC7A is
      unable to check the call level for nested interrupts. */

 The compiler will detect if the initially mentioned registers are
 modified during interrupt processing without being saved and
 restored. The supplied macros for saving and restoring registers
 will only save W, STATUS and BSR and PCLATH. The other registers
 have to be saved and restored by user code when needed.

 For example, if FSR0 is modified by a table or pointer access, or
 by direct writing, the compiler will check that FSR0 is saved and
 restored, also in nested function calls. Note that the FSR0 saving
 and restoring can be done in a local region surrounding the indexed
 access, and does not need to be done in the beginning and end of
 the interrupt routine.

 A warning is printed if the Group 2 - 4 registers mentioned above
 are saved but not changed.


 The error and warning messages printed can be removed:

  #pragma interruptSaveCheck  n  // no warning or error
  #pragma interruptSaveCheck  w  // warning only
  #pragma interruptSaveCheck  e  // error and warning (default)

 Note that the above pragma also change the checking done on all
 registers.


// CUSTOM INTERRUPT SAVE AND RESTORE:

 It is not required to use the above save and restore macros. CC7A
 also supports custom interrupt structures.

  A) You may want to use your own save and restore sequence. This
     can be done by inline assembly. Take a look at the macros at
     the end of this file to get an idea on how to do this. If CC7A
     does not accept your code, just insert (on your own risk):

       #pragma interruptSaveCheck  n  // no warning or error

  B) No registers need to be saved when using the following
     instructions in the interrupt routine. The register save
     checking should NOT be disabled.

       bx2 = 1;        // BSF   0x1F,bx2  ; unbanked locations only
       bx1 = 0;        // BCF   0x1F,bx1  ; unbanked locations only
       bx3 = !bx3;     // BTG   0x1F,bx3  ; unbanked locations only
       btss(bx1);      // BTFSS 0x1F,bx1  ; unbanked locations only
       btsc(bx1);      // BTFSC 0x1F,bx1  ; unbanked locations only
       vs = swap(vs);  // SWAPF vs,1      ; unbanked locations only
       vs = incsz(vs); // INCFSZ vs,1     ; unbanked locations only
       vs = decsz(vs); // DECFSZ vs,1     ; unbanked locations only
       vs = incsnz(vs);// INFSNZ vs,1     ; unbanked locations only
       vs = decsnz(vs);// DCFSNZ vs,1     ; unbanked locations only
       a = b;          // MOVFP a,b       ; unbanked locations only
       a = 0;          // CLRF a,1        ; unbanked locations only
       a = 0xFF;       // SETF a,1        ; unbanked locations only
       vs = rrnc(vs);  // RRNC vs,1       ; unbanked locations only
       vs = rlnc(vs);  // RRNC vs,1       ; unbanked locations only
       skipIfZero(vs); // TSTFSZ vs       ; unbanked locations only
       nop();          // NOP
       clrwdt();       // CLRWDT

  C) It is possible to enable interrupt only in special regions
     (wait loops) in such a way that main registers can
     be modified during interrupt without disturbing the main
     program. The register save can then be omitted and the save
     checking must be swithed off to avoid error messages:

       #pragma interruptSaveCheck  n  // no warning or error
*/


 #define  multi_interrupt_entry_and_save \
    shrBank char sBSR, sALUSTA; \
    bank0 char sWREG, sPCLATH; \
        \
   #asm \
    ; ADDRESS 0x08 \
    MOVFP BSR,sBSR           ; unbanked \
    MOVFP ALUSTA, sALUSTA    ; unbanked \
    CLRF BSR \
    MOVPF PCLATH, sPCLATH    ; bank 0 \
    MOVWF sWREG              ; bank 0 \
    CLRF PCLATH \
    GOTO INT_service \
    NOP \
         \
    ; ADDRESS 0x10 \
    MOVFP BSR,sBSR           ; unbanked \
    MOVFP ALUSTA, sALUSTA    ; unbanked \
    CLRF BSR \
    MOVPF PCLATH, sPCLATH    ; bank 0 \
    MOVWF sWREG              ; bank 0 \
    CLRF PCLATH \
    GOTO TMR0_service \
    NOP \
        \
    ; ADDRESS 0x18 \
    MOVFP BSR,sBSR           ; unbanked \
    MOVFP ALUSTA, sALUSTA    ; unbanked \
    CLRF BSR \
    MOVPF PCLATH, sPCLATH    ; bank 0 \
    MOVWF sWREG              ; bank 0 \
    CLRF PCLATH \
    GOTO T0CKI_service \
    NOP \
        \
    ; ADDRESS 0x20 \
    MOVFP BSR,sBSR           ; unbanked \
    MOVFP ALUSTA, sALUSTA    ; unbanked \
    CLRF BSR \
    MOVPF PCLATH, sPCLATH    ; bank 0 \
    MOVWF sWREG              ; bank 0 \
    CLRF PCLATH \
   #endasm \
    goto PERIPHERAL_service; \



 /* Next alternative uses only one unbanked location, but need one instruction
    more to save vital registers compared to the above macro */

 #define  multi_interrupt_entry_and_save_alt2 \
    shrBank char sBSR; \
    bank0 char sWREG, sPCLATH, sALUSTA; \
        \
   #asm \
    ; ADDRESS 0x08 \
    MOVFP BSR,sBSR           ; unbanked \
    CLRF BSR \
    MOVWF sWREG              ; bank 0 \
    MOVFP ALUSTA,WREG                 \
    MOVWF sALUSTA            ; bank 0 \
    MOVPF PCLATH, sPCLATH    ; bank 0 \
    CLRF PCLATH \
    GOTO INT_service \
         \
    ; ADDRESS 0x10 \
    MOVFP BSR,sBSR           ; unbanked \
    CLRF BSR \
    MOVWF sWREG              ; bank 0 \
    MOVFP ALUSTA,WREG                 \
    MOVWF sALUSTA            ; bank 0 \
    MOVPF PCLATH, sPCLATH    ; bank 0 \
    CLRF PCLATH \
    GOTO TMR0_service \
        \
    ; ADDRESS 0x18 \
    MOVFP BSR,sBSR           ; unbanked \
    CLRF BSR \
    MOVWF sWREG              ; bank 0 \
    MOVFP ALUSTA,WREG                 \
    MOVWF sALUSTA            ; bank 0 \
    MOVPF PCLATH, sPCLATH    ; bank 0 \
    CLRF PCLATH \
    GOTO T0CKI_service \
        \
    ; ADDRESS 0x20 \
    MOVFP BSR,sBSR           ; unbanked \
    CLRF BSR \
    MOVWF sWREG              ; bank 0 \
    MOVFP ALUSTA,WREG                 \
    MOVWF sALUSTA            ; bank 0 \
    MOVPF PCLATH, sPCLATH    ; bank 0 \
    CLRF PCLATH \
   #endasm \
    goto PERIPHERAL_service; \





 #define interrupt_exit_and_restore \
   #asm \
    MOVFP sWREG,W            \
    MOVFP sPCLATH, PCLATH    \
    MOVFP sALUSTA, ALUSTA    \
    MOVFP sBSR, BSR          \
    RETFIE \
   #endasm \
